#ifndef __ENET_PROTOCOL_H__
#define __ENET_PROTOCOL_H__

#include "types.h"

enum
{
   ENET_PROTOCOL_MINIMUM_PACKET_SIZE     = 576,
   ENET_PROTOCOL_MAXIMUM_PACKET_SIZE     = 4096,
   ENET_PROTOCOL_MAXIMUM_PACKET_COMMANDS = 32,
   ENET_PROTOCOL_MINIMUM_WINDOW_SIZE     = 4096,
   ENET_PROTOCOL_MAXIMUM_WINDOW_SIZE     = 32768,
   ENET_PROTOCOL_MINIMUM_CHANNEL_COUNT   = 1,
   ENET_PROTOCOL_MAXIMUM_CHANNEL_COUNT   = 255
};

typedef enum
{
   ENET_PROTOCOL_COMMAND_NONE               = 0,
   ENET_PROTOCOL_COMMAND_ACKNOWLEDGE        = 1,
   ENET_PROTOCOL_COMMAND_CONNECT            = 2,
   ENET_PROTOCOL_COMMAND_VERIFY_CONNECT     = 3,
   ENET_PROTOCOL_COMMAND_DISCONNECT         = 4,
   ENET_PROTOCOL_COMMAND_PING               = 5,
   ENET_PROTOCOL_COMMAND_SEND_RELIABLE      = 6,
   ENET_PROTOCOL_COMMAND_SEND_UNRELIABLE    = 7,
   ENET_PROTOCOL_COMMAND_SEND_FRAGMENT      = 8,
   ENET_PROTOCOL_COMMAND_BANDWIDTH_LIMIT    = 9,
   ENET_PROTOCOL_COMMAND_THROTTLE_CONFIGURE = 10
} ENetProtocolCommand;

typedef enum
{
   ENET_PROTOCOL_FLAG_ACKNOWLEDGE = (1 << 0)
} ENetProtocolFlag;

typedef struct
{
   uint16 peerID;
   uint8 flags;
   uint8 commandCount;
   uint32 sentTime;
   uint32 challenge;
} ENetProtocolHeader;

typedef struct
{
   uint8 command;
   uint8 channelID;
   uint8 flags;
   uint8 reserved;
   uint32 commandLength;
   uint32 reliableSequenceNumber;
} ENetProtocolCommandHeader;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint32 receivedReliableSequenceNumber;
   uint32 receivedSentTime;
} ENetProtocolAcknowledge;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint16 outgoingPeerID;
   uint16 packetSize;
   uint32 windowSize;
   uint32 channelCount;
   uint32 incomingBandwidth;
   uint32 outgoingBandwidth;
   uint32 packetThrottleInterval;
   uint32 packetThrottleAcceleration;
   uint32 packetThrottleDeceleration;
} ENetProtocolConnect;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint16 outgoingPeerID;
   uint16 packetSize;
   uint32 windowSize;
   uint32 channelCount;
   uint32 incomingBandwidth;
   uint32 outgoingBandwidth;
   uint32 packetThrottleInterval;
   uint32 packetThrottleAcceleration;
   uint32 packetThrottleDeceleration;
} ENetProtocolVerifyConnect;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint32 incomingBandwidth;
   uint32 outgoingBandwidth;
} ENetProtocolBandwidthLimit;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint32 packetThrottleInterval;
   uint32 packetThrottleAcceleration;
   uint32 packetThrottleDeceleration;
} ENetProtocolThrottleConfigure;

typedef struct
{
   ENetProtocolCommandHeader header;
} ENetProtocolDisconnect;

typedef struct
{
   ENetProtocolCommandHeader header;
} ENetProtocolPing;

typedef struct
{
   ENetProtocolCommandHeader header;
} ENetProtocolSendReliable;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint32 unreliableSequenceNumber;
} ENetProtocolSendUnreliable;

typedef struct
{
   ENetProtocolCommandHeader header;
   uint32 startSequenceNumber;
   uint32 fragmentCount;
   uint32 fragmentNumber;
   uint32 totalLength;
   uint32 fragmentOffset;
} ENetProtocolSendFragment;

typedef union
{
   ENetProtocolCommandHeader header;
   ENetProtocolAcknowledge acknowledge;
   ENetProtocolConnect connect;
   ENetProtocolVerifyConnect verifyConnect;
   ENetProtocolDisconnect disconnect;
   ENetProtocolPing ping;
   ENetProtocolSendReliable sendReliable;
   ENetProtocolSendUnreliable sendUnreliable;
   ENetProtocolSendFragment sendFragment;
   ENetProtocolBandwidthLimit bandwidthLimit;
   ENetProtocolThrottleConfigure throttleConfigure;
} ENetProtocol;

#endif /* __ENET_PROTOCOL_H__ */

