#ifdef WIN32

#include <time.h>
#include <enet/enet.h>

static uint32 timeBase = 0;

int
enet_initialize (void)
{
    WORD versionRequested = MAKEWORD (1, 1);
    WSADATA wsaData;
   
    if (WSAStartup (versionRequested, & wsaData))
       return -1;

    if (LOBYTE (wsaData.wVersion) != 1||
        HIBYTE (wsaData.wVersion) != 1)
    {
       WSACleanup ();
       
       return -1;
    }

    return 0;
}

void
enet_deinitialize (void)
{
    WSACleanup ();
}

uint32
enet_time_get (void)
{
    return (uint32) GetTickCount () - timeBase;
}

uint32
enet_time_get_sec (void)
{

    return (uint32) (GetTickCount () - timeBase) / 1000;  // FIXME: get higher res time from somewhere
    /*
    SYSTEMTIME st;
    GetSystemTime(&st);
    FILETIME ft;
    assert(SystemTimeToFileTime(&st, &ft));
    return (uint32) (((ft.dwLowDateTime>>23)|(ft.dwHighDateTime<<9))/(10000000/(float)1>>23));
    */
}

void
enet_time_set (uint32 newTimeBase)
{
    timeBase = (uint32) GetTickCount () - newTimeBase;
}

int
enet_address_set_host (ENetAddress * address, const char * name)
{
    struct hostent * hostEntry;

    hostEntry = gethostbyname (name);
    if (hostEntry == NULL ||
        hostEntry -> h_addrtype != AF_INET)
      return -1;

    address -> host = * (uint32 *) hostEntry -> h_addr_list [0];

    return 0;
}

int
enet_address_get_host (const ENetAddress * address, char * name, size_t nameLength)
{
    struct in_addr in;
    struct hostent * hostEntry;
    
    in.s_addr = address -> host;
    
    hostEntry = gethostbyaddr ((char *) & in, sizeof (struct in_addr), AF_INET);
    if (hostEntry == NULL)
      return -1;

    strncpy (name, hostEntry -> h_name, nameLength);

    return 0;
}

ENetSocket 
enet_socket_create (ENetSocketType type, const ENetAddress * address) 
{ 
    ENetSocket newSocket = socket (PF_INET, type == ENET_SOCKET_TYPE_DATAGRAM ? SOCK_DGRAM : SOCK_STREAM, 0); 
    int nonBlocking = 1, 
        receiveBufferSize = ENET_HOST_RECEIVE_BUFFER_SIZE; 
    struct sockaddr_in sin; 

    if (newSocket == ENET_SOCKET_NULL) 
      return ENET_SOCKET_NULL; 

    if (type == ENET_SOCKET_TYPE_DATAGRAM) 
    { 
        ioctlsocket (newSocket, FIONBIO, & nonBlocking); 

        setsockopt (newSocket, SOL_SOCKET, SO_RCVBUF, (char *) & receiveBufferSize, sizeof (int)); 
    } 

    memset (& sin, 0, sizeof (struct sockaddr_in)); 

    sin.sin_family = AF_INET; 

    if (address != NULL) 
    { 
       sin.sin_port = ENET_HOST_TO_NET_16 (address -> port); 
       sin.sin_addr.s_addr = address -> host; 
    } 
    else 
    { 
/* need to bind client sockets to a port, since windows will otherwise 
* only do the binding on the first socket send 
*/ 
       sin.sin_port = 0; 
       sin.sin_addr.s_addr = INADDR_ANY; 
    } 

    if ((type == ENET_SOCKET_TYPE_STREAM && 
/* the following check makes sure that if address == NULL (i.e. a client socket) 
* that the listen never gets called, but it still falls through to the bind 
* so that the client socket can be bound immediately to port 
*/ 
          address != NULL && 
          listen (newSocket, SOMAXCONN) == SOCKET_ERROR) || 
        bind (newSocket, 
              (struct sockaddr *) & sin, 
              sizeof (struct sockaddr_in)) == SOCKET_ERROR) 
    { 
       closesocket (newSocket); 
    
       return ENET_SOCKET_NULL; 
    } 

    return newSocket; 
} 

/*
ENetSocket
enet_socket_create (ENetSocketType type, const ENetAddress * address)
{
    ENetSocket newSocket = socket (PF_INET, type == ENET_SOCKET_TYPE_DATAGRAM ? SOCK_DGRAM : SOCK_STREAM, 0);
    int nonBlocking = 1,
        receiveBufferSize = ENET_HOST_RECEIVE_BUFFER_SIZE;
    struct sockaddr_in sin;

    if (newSocket == ENET_SOCKET_NULL)
      return ENET_SOCKET_NULL;

    if (type == ENET_SOCKET_TYPE_DATAGRAM)
    {
        ioctlsocket (newSocket, FIONBIO, & nonBlocking);

        setsockopt (newSocket, SOL_SOCKET, SO_RCVBUF, (char *) & receiveBufferSize, sizeof (int));
    }

    memset (& sin, 0, sizeof (struct sockaddr_in));

    sin.sin_family = AF_INET;
    
    if (address != NULL)
    {
       sin.sin_port = ENET_HOST_TO_NET_16 (address -> port);
       sin.sin_addr.s_addr = address -> host;
    }
    else
    {
       sin.sin_port = 0;
       sin.sin_addr.s_addr = INADDR_ANY;
    }

    if ((type == ENET_SOCKET_TYPE_STREAM &&
          listen (newSocket, SOMAXCONN) == SOCKET_ERROR) ||
        bind (newSocket,    
              (struct sockaddr *) & sin,
              sizeof (struct sockaddr_in)) == SOCKET_ERROR)
    {
       closesocket (newSocket);

       return ENET_SOCKET_NULL;
    }

    return newSocket;
}
*/

int
enet_socket_connect (ENetSocket socket, const ENetAddress * address)
{
    struct sockaddr_in sin;

    memset (& sin, 0, sizeof (struct sockaddr_in));

    sin.sin_family = AF_INET;
    sin.sin_port = ENET_HOST_TO_NET_16 (address -> port);
    sin.sin_addr.s_addr = address -> host;

    return connect (socket, (struct sockaddr *) & sin, sizeof (struct sockaddr_in));
}

ENetSocket
enet_socket_accept (ENetSocket socket, ENetAddress * address)
{
    int result;
    struct sockaddr_in sin;
    int sinLength = sizeof (struct sockaddr_in);

    result = accept (socket, 
                     address != NULL ? (struct sockaddr *) & sin : NULL, 
                     address != NULL ? & sinLength : NULL);

    if (result == -1)
      return ENET_SOCKET_NULL;

    if (address != NULL)
    {
        address -> host = (uint32) sin.sin_addr.s_addr;
        address -> port = ENET_NET_TO_HOST_16 (sin.sin_port);
    }

    return result;
}

void
enet_socket_destroy (ENetSocket socket)
{
    closesocket (socket);
}

int
enet_socket_send (ENetSocket socket,
                  const ENetAddress * address,
                  const ENetBuffer * buffers,
                  size_t bufferCount)
{
    struct sockaddr_in sin;
    DWORD sentLength;

    if (address != NULL)
    {
        sin.sin_family = AF_INET;
        sin.sin_port = ENET_HOST_TO_NET_16 (address -> port);
        sin.sin_addr.s_addr = address -> host;
    }

    if (WSASendTo (socket, 
                   (LPWSABUF) buffers,
                   (DWORD) bufferCount,
                   & sentLength,
                   0,
                   address != NULL ? (struct sockaddr *) & sin : 0,
                   address != NULL ? sizeof (struct sockaddr_in) : 0,
                   NULL,
                   NULL) == SOCKET_ERROR)
    {
       if (WSAGetLastError () == WSAEWOULDBLOCK)
         return 0;

       return -1;
    }

    return (int) sentLength;
}

int
enet_socket_receive (ENetSocket socket,
                     ENetAddress * address,
                     ENetBuffer * buffers,
                     size_t bufferCount)
{
    DWORD sinLength = sizeof (struct sockaddr_in),
          flags = 0,
          recvLength;
    struct sockaddr_in sin;

    if (WSARecvFrom (socket,
                     (LPWSABUF) buffers,
                     (DWORD) bufferCount,
                     & recvLength,
                     & flags,
                     address != NULL ? (struct sockaddr *) & sin : NULL,
                     address != NULL ? & sinLength : NULL,
                     NULL,
                     NULL) == SOCKET_ERROR)
    {
       switch (WSAGetLastError ())
       {
       case WSAEWOULDBLOCK:
       case WSAECONNRESET:
          return 0;
       }

       return -1;
    }

    if (flags & MSG_PARTIAL)
      return -1;

    if (address != NULL)
    {
        address -> host = (uint32) sin.sin_addr.s_addr;
        address -> port = ENET_NET_TO_HOST_16 (sin.sin_port);
    }

    return (int) recvLength;
}

int
enet_socket_wait (ENetSocket socket, uint32 * condition, uint32 timeout)
{
    fd_set readSet, writeSet;
    struct timeval timeVal;
    int selectCount;
    
    timeVal.tv_sec = timeout / 1000;
    timeVal.tv_usec = (timeout % 1000) * 1000;
    
    FD_ZERO (& readSet);
    FD_ZERO (& writeSet);

    if (* condition & ENET_SOCKET_WAIT_SEND)
      FD_SET (socket, & writeSet);

    if (* condition & ENET_SOCKET_WAIT_RECEIVE)
      FD_SET (socket, & readSet);

    selectCount = select (socket + 1, & readSet, & writeSet, NULL, & timeVal);

    if (selectCount < 0)
      return -1;

    * condition = ENET_SOCKET_WAIT_NONE;

    if (selectCount == 0)
      return 0;

    if (FD_ISSET (socket, & writeSet))
      * condition |= ENET_SOCKET_WAIT_SEND;
    
    if (FD_ISSET (socket, & readSet))
      * condition |= ENET_SOCKET_WAIT_RECEIVE;

    return 0;
} 

#endif

